import { __rest } from "tslib";
import { normalizeClassName } from 'petals-ui/dist/basic';
import { isBoolean, isNumber, isString, isNumeric, isFunction, isPlainObject, omit, } from '../../vendors/toolbox';
import { getControl, getViews, createModuleContext, createView, } from '../../vendors/organik';
import { getAppHelperForInternalUse } from '../../vendors/organik/core/app';
import { isBoolOrExprTrue } from '../renderer';
import { getBehaviorByKey } from '../theme';
import { resolveSafePropsFromConfig } from './base';
const BREAKPOINT_PART = '(xs|sm|md|lg|xl)-[1-9]{1,2}';
const fullRegExp = new RegExp(`^([?${BREAKPOINT_PART}(/${BREAKPOINT_PART}){0,}]?,?)+$`);
function resolveBreakpoints(cols) {
    return cols.split(',').map(col => col
        .replace(/(\[|\])/g, '')
        .split('/')
        .reduce((prev, breakpoint) => {
        const [size, span] = breakpoint.split('-');
        return Object.assign(Object.assign({}, prev), { [size]: Number(span) });
    }, {}));
}
function renderFormFieldNodes(descriptors, arrangement, renderChild, renderRow) {
    const formFieldNodes = [];
    const rows = (arrangement || '').split('|');
    const needLayout = rows.length > 0 &&
        rows.filter(row => (isNumeric(row) && Number(row) > 0) || fullRegExp.test(row)).length ===
            rows.length;
    if (needLayout) {
        const remainedFilters = descriptors.filter(({ hidden }) => hidden !== true);
        do {
            const cols = rows.shift();
            let breakpoints = [];
            let count;
            if (isNumeric(cols)) {
                count = Number(cols);
            }
            else {
                breakpoints = resolveBreakpoints(cols);
                count = breakpoints.length;
            }
            formFieldNodes.push(renderRow(remainedFilters.splice(0, count), breakpoints));
        } while (remainedFilters.length > 0 && rows.length > 0);
        if (remainedFilters.length > 0) {
            formFieldNodes.push(renderRow(remainedFilters, -1));
        }
    }
    else {
        descriptors.forEach(descriptor => {
            if (!descriptor.hidden) {
                formFieldNodes.push(renderChild(descriptor));
            }
        });
    }
    return formFieldNodes;
}
function renderFormChildren(descriptors, arrangement, renderGroup, renderChild, renderRow) {
    const groups = (arrangement || '').match(/\(([^()]+)\)/g);
    if (!groups) {
        return renderFormFieldNodes(descriptors, arrangement, renderChild, renderRow);
    }
    let startIndex = 0;
    let endIndex = 0;
    return groups
        .map(group => group.replace(/\(|\)/g, ''))
        .filter(group => !!group)
        .map(group => {
        const grouped = group.split(':');
        const count = grouped[1]
            .split('|')
            .reduce((prev, row) => (isNumeric(row) ? Number(row) : row.split(',').length) + prev, 0);
        endIndex = startIndex + count;
        const fields = renderGroup(grouped[0], renderFormFieldNodes(descriptors.slice(startIndex, endIndex), grouped[1], renderChild, renderRow));
        startIndex = endIndex;
        return fields;
    });
}
function resolveFormLayout({ config, behavior }) {
    if (config.formLayout) {
        return config.formLayout;
    }
    return behavior && behavior.formLayout
        ? behavior.formLayout
        : getBehaviorByKey('common.view.objectViewFormLayout', 'horizontal');
}
function resolveLabelWidth({ formControlLabelWidth }) {
    const labelWidth = formControlLabelWidth || getBehaviorByKey('common.view.objectViewFormControlLabelWidth', '');
    return isNumber(labelWidth) ? `${labelWidth}px` : labelWidth;
}
function resolveFormControlSize({ formControlSize, }) {
    return formControlSize || getBehaviorByKey('common.view.objectViewFormControlSize', '');
}
function resolveShowValidationMessage({ showValidationMessage }) {
    return isBoolean(showValidationMessage)
        ? showValidationMessage
        : getBehaviorByKey('common.view.objectViewShowValidationMessage');
}
function renderForm(props, renderGroup, renderChild, renderRow, render) {
    const { value, config = {} } = props;
    const { arrangement } = config;
    return render({
        FormControl: getControl('Form'),
        props: {
            className: normalizeClassName('HandieFormRenderer', props.className),
            layout: resolveFormLayout(props),
            labelOption: { width: resolveLabelWidth(config), align: 'right' },
            controlSize: resolveFormControlSize(config),
            hideMessage: !resolveShowValidationMessage(config),
        },
        children: renderFormChildren(props.fields.filter(({ available }) => !isString(available) || isBoolOrExprTrue(value, available, false)), isFunction(arrangement) ? arrangement(Object.assign({}, (value || {}))) : arrangement, renderGroup, renderChild, renderRow),
    });
}
function resolveViewWidgetCtor(props, componentAsserter, moduleContextCreator = createModuleContext) {
    const [moduleName, resourceType, viewName] = props.view.split('.');
    const view = getViews(moduleName)[viewName];
    if (resourceType !== 'views' || !view) {
        return undefined;
    }
    return createView(moduleContextCreator(moduleName), isFunction(view) && !componentAsserter(view)
        ? view(...props.params)
        : view);
}
function isConditionPersisted(viewContext) {
    const defaultConditionPersists = getBehaviorByKey('common.search.conditionPersists', false);
    const searchDescriptor = viewContext.getSearch();
    if (!isPlainObject(searchDescriptor)) {
        return defaultConditionPersists;
    }
    const { conditionPersists } = searchDescriptor.config || {};
    return isBoolean(conditionPersists) ? conditionPersists : defaultConditionPersists;
}
function encode(condition) {
    return condition ? encodeURIComponent(JSON.stringify(condition)) : '';
}
function decode(encoded) {
    return encoded ? JSON.parse(decodeURIComponent(encoded)) : {};
}
const persistenceKey = '__handie-scpk'; // !!! can't be dynamic, because URL will be shared among people
function persistsInUrl(history, params) {
    const _a = omit(history.getLocation(), ['rawPath', 'ancestors']), { query } = _a, others = __rest(_a, ["query"]);
    history.replace(Object.assign(Object.assign({}, others), { query: Object.assign(Object.assign({}, query), { [persistenceKey]: encode(Object.assign(Object.assign({}, decode(query[persistenceKey])), params)) }) }));
}
function restoreFromUrl(history) {
    return decode(history.getLocation().query[persistenceKey]);
}
function isActionsAuthorized(actionsAuthority) {
    if (!actionsAuthority) {
        return true;
    }
    const authority = getAppHelperForInternalUse().session.getAuthority();
    return authority ? !!authority[actionsAuthority] : false;
}
function resolveActionsAuthority(viewContext) {
    const actionsAuthority = viewContext.getActionsAuthority();
    const { history, session } = getAppHelperForInternalUse();
    return isFunction(actionsAuthority)
        ? actionsAuthority(history, session)
        : actionsAuthority;
}
function resolveAuthorizedActions(actions, actionsAuthority) {
    if (actionsAuthority) {
        return actions;
    }
    const { history, session } = getAppHelperForInternalUse();
    const authority = session.getAuthority();
    if (!authority) {
        return [];
    }
    return actions.filter(({ authority: auth }) => {
        const resolvedAuth = isFunction(auth)
            ? auth(history, session)
            : auth;
        return !resolvedAuth || !!authority[resolvedAuth];
    });
}
function resolveAuthorizedSpecificActions(actionFilter, viewContext) {
    const actionsAuthority = resolveActionsAuthority(viewContext);
    return isActionsAuthorized(actionsAuthority)
        ? resolveAuthorizedActions(viewContext.getActions().filter(actionFilter), actionsAuthority)
        : [];
}
const resolveTopActions = resolveAuthorizedSpecificActions.bind(null, ({ context }) => !!context && context !== 'single');
const resolveItemActions = resolveAuthorizedSpecificActions.bind(null, ({ context }) => !context || context === 'single');
function resolveActionStuffs(viewContext) {
    const topActions = [];
    const itemActions = [];
    const actionMap = {};
    const actionsAuthority = resolveActionsAuthority(viewContext);
    if (isActionsAuthorized(actionsAuthority)) {
        resolveAuthorizedActions(viewContext.getActions(), actionsAuthority).forEach(action => {
            if (!action.context || action.context === 'single') {
                itemActions.push(action);
            }
            else {
                topActions.push(action);
            }
            if (action.name) {
                actionMap[action.name] = action;
            }
        });
    }
    return { top: topActions, item: itemActions, map: actionMap };
}
function resolveTooltipEnabled(field, context) {
    if (field.config && isBoolean(field.config.showOverflowTooltip)) {
        return field.config.showOverflowTooltip;
    }
    const { showTooltipWhenContentOverflow } = context.getConfig();
    return isBoolean(showTooltipWhenContentOverflow)
        ? showTooltipWhenContentOverflow
        : getBehaviorByKey('view.table.showTooltipWhenContentOverflow', false);
}
function resolveTableColumns(context, inlineButtonSize, inlineActionRenderType, cellRendererResolver, operationColumnResolver) {
    const cols = context.getFields().map(field => (Object.assign({ key: field.name, title: field.label, ellipsis: resolveTooltipEnabled(field, context), render: cellRendererResolver(field, context) }, (field.config || {}))));
    const { showSerialNumber, serialNumberColumnWidth = getBehaviorByKey('view.table.serialNumberColumnWidth', '55'), serialNumberColumnAlignment = 'center', checkable, selectionColumnWidth = getBehaviorByKey('view.table.selectionColumnWidth', '55'), selectionColumnAlignment = 'center', } = context.getConfig();
    if (showSerialNumber === true) {
        cols.unshift({
            title: '序号',
            type: 'index',
            width: serialNumberColumnWidth,
            align: serialNumberColumnAlignment,
        });
    }
    const actionsAuthority = resolveActionsAuthority(context);
    const checkableActions = isActionsAuthorized(actionsAuthority)
        ? resolveAuthorizedActions([].concat((context.getActionsByContextType('batch') || []), (context.getActionsByContextType('both') || [])), actionsAuthority)
        : [];
    if (isBoolean(checkable) ? checkable : checkableActions.length > 0) {
        cols.unshift({
            type: 'selection',
            width: selectionColumnWidth,
            align: selectionColumnAlignment,
        });
    }
    const operationCol = operationColumnResolver(context, inlineButtonSize, inlineActionRenderType);
    if (operationCol) {
        cols.push(operationCol);
    }
    return cols;
}
function resolveTableProps(context, inlineButtonSize, inlineActionRenderType, cellRendererResolver, operationColumnResolver) {
    return Object.assign(Object.assign({}, resolveSafePropsFromConfig(context.getConfig(), [
        'checkable',
        'showSerialNumber',
        'showTooltipWhenContentOverflow',
        'selectionColumnWidth',
        'selectionColumnAlignment',
        'serialNumberColumnWidth',
        'serialNumberColumnAlignment',
        'operationColumnWidth',
        'operationColumnAlignment',
    ])), { columns: resolveTableColumns(context, inlineButtonSize, inlineActionRenderType, cellRendererResolver, operationColumnResolver) });
}
function resolveSafeDialogProps(config) {
    return resolveSafePropsFromConfig(config, [
        'width',
        'visible',
        'affirmButton',
        'denyButton',
        'onClose',
    ]);
}
export { renderFormChildren, renderForm, resolveViewWidgetCtor, isConditionPersisted, persistsInUrl, restoreFromUrl, isActionsAuthorized, resolveActionsAuthority, resolveAuthorizedActions, resolveItemActions, resolveActionStuffs, resolveTableProps, resolveSafeDialogProps, };
